# Copyright (c) 2008 gocept gmbh & co. kg
# See also LICENSE.txt

from zope.testing import doctest
import gocept.runner
import os.path
import unittest
import zope.app.appsetup.product
import zope.app.testing.functional


runner_layer = zope.app.testing.functional.ZCMLLayer(
    os.path.join(os.path.dirname(__file__), 'ftesting.zcml'),
    __name__, 'RunnerLayer', allow_teardown=True)


class TestFromConfig(unittest.TestCase):

    def setUp(self):
        self.old_config = zope.app.appsetup.product.saveConfiguration()
        zope.app.appsetup.product.setProductConfiguration('test', dict(
            foo='bar'))

    def tearDown(self):
        zope.app.appsetup.product.restoreConfiguration(self.old_config)

    def test_invalid_section(self):
        self.assertRaises(KeyError,
                          gocept.runner.from_config('invalid', 'foo'))

    def test_invalid_value(self):
        self.assertRaises(KeyError,
                          gocept.runner.from_config('test', 'invalid'))

    def test_valid(self):
        self.assertEqual('bar',
                         gocept.runner.from_config('test', 'foo')())

    def test_returns_callable(self):
        self.assertTrue(callable(gocept.runner.from_config('baz', 'boink')))


flags = doctest.INTERPRET_FOOTNOTES | doctest.ELLIPSIS


def test_suite():
    suite = unittest.TestSuite()
    test = zope.app.testing.functional.FunctionalDocFileSuite(
        'README.txt',
        optionflags=flags)
    test.layer = runner_layer
    suite.addTest(test)

    suite.addTest(doctest.DocFileSuite(
        'appmain.txt',
        'once.txt',
        optionflags=flags))
    suite.addTest(unittest.makeSuite(TestFromConfig))

    return suite



# Copyright (c) 2008 gocept gmbh & co. kg
# See also LICENSE.txt
"""Infrastructure for running."""

import logging
import os
import time
import signal

import ZODB.POSException
import transaction
import zope.app.appsetup.product
import zope.app.security.interfaces
import zope.app.component.hooks
import zope.app.server.main
import zope.app.wsgi
import zope.security.management
import zope.security.testing
import zope.publisher.base


log = logging.getLogger(__name__)


Exit = object()


class RunnerRequest(zope.publisher.base.BaseRequest):
    """A custom publisher request for the runner."""

    def __init__(self, *args):
        super(RunnerRequest, self).__init__(None, {}, positional=args)


class MainLoop(object):

    def __init__(self, app, ticks, worker, principal=None, once=False):
        self._is_running = False
        self.app = app
        self.ticks = ticks
        self.worker = worker
        self.once = once
        if principal is None:
            self.interaction = False
        else:
            self.interaction = True
            self.principal_id = principal

    def stopMainLoop(self, signum, frame):
        log.info("Received signal %s, terminating." % signum)
        self._is_running = False

    def __call__(self):
        old_site = zope.app.component.hooks.getSite()
        zope.app.component.hooks.setSite(self.app)

        self._is_running = True

        while self._is_running:
            ticks = None
            self.begin()
            try:
                ticks = self.worker()
            except (KeyboardInterrupt, SystemExit):
                self.abort()
                break
            except Exception, e:
                log.exception(e)
                self.abort()
            else:
                try:
                    self.commit()
                except ZODB.POSException.ConflictError, e:
                    log.exception(e)
                    self.abort()
                    # Silently ignore this. The next run will be a retry 
                    # anyways.

            if self.once or ticks is Exit:
                self._is_running = False
            else:
                if ticks is None:
                    ticks = self.ticks
                log.debug("Sleeping %s seconds" % ticks)
                time.sleep(ticks)

        zope.app.component.hooks.setSite(old_site)

    def begin(self):
        transaction.begin()
        if self.interaction:
            request = RunnerRequest()
            request.setPrincipal(self.principal)
            zope.security.management.newInteraction(request)

    def abort(self):
        transaction.abort()
        if self.interaction:
            zope.security.management.endInteraction()

    def commit(self):
        transaction.commit()
        if self.interaction:
            zope.security.management.endInteraction()

    @property
    def principal(self):
        auth = zope.component.getUtility(
            zope.app.security.interfaces.IAuthentication)
        return auth.getPrincipal(self.principal_id)


class appmain(object):
    """Decorator to simplify the actual entry point functions for main loops.
    """

    def __init__(self, ticks=1, principal=None):
        self.ticks = ticks
        self.principal = principal
        self.once = False

    def get_principal(self):
        if callable(self.principal):
            return self.principal()
        # BBB: depcreated:
        return self.principal

    def __call__(self, worker_method):
        def configure(appname, configfile):
            db, app = init(appname, configfile)
            mloop = MainLoop(app, self.ticks, worker_method,
                             principal=self.get_principal(),
                             once=self.once)
            # XXX do we want more signal handlers?
            signal.signal(signal.SIGHUP, mloop.stopMainLoop)
            signal.signal(signal.SIGTERM, mloop.stopMainLoop)
            mloop()
            db.close()
        # Just to make doctests look nice.
        configure.__name__ = worker_method.__name__
        return configure


class once(appmain):

    def __init__(self, principal=None):
        super(once, self).__init__(principal=principal)
        self.once = True


def init(appname, configfile):
    """Initialise the Zope environment (without network servers) and return a
    specific root-level object.
    """
    options = zope.app.server.main.load_options(['-C', configfile])
    zope.app.appsetup.product.setProductConfigurations(options.product_config)

    db = zope.app.wsgi.config(
        configfile,
        schemafile=os.path.join(
            os.path.dirname(zope.app.server.main.__file__), 'schema.xml'))

    root = db.open().root()
    app = root['Application']
    if appname is not None:
        app = app[appname]
    return db, app


def from_config(section, variable):
    """Return a callable that returns a value from product config."""
    def get_config():
        config = zope.app.appsetup.product.getProductConfiguration(section)
        if config is None:
            raise KeyError('No such section %s' % section)
        return config[variable]
    return get_config



# Copyright (c) 2008 gocept gmbh & co. kg
# See also LICENSE.txt
"""Example to demonstrate how to create a "run" script."""

import gocept.runner


@gocept.runner.appmain(ticks=1)
def example():
    """Example function. This is called every ticks secons."""
    print "Calculating..."



import os.path
from setuptools import setup, find_packages

def read(*names):
    return open(os.path.join(os.path.dirname(__file__), *names), 'r').read()

setup(name='gocept.runner',
      version = '0.5.1',
      description="Create stand alone programs with full ZCA",
      long_description=(
          read('src', 'gocept', 'runner', 'appmain.txt')
          + '\n\n'
          + read('src', 'gocept', 'runner', 'once.txt')
          + '\n\n'
          + read('src', 'gocept', 'runner', 'README.txt')
          + '\n\n'
          + read('CHANGES.txt')
      ),
      # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers = (
          "Topic :: Software Development",
          "Framework :: Zope3",
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved",
          "License :: OSI Approved :: Zope Public License",
          "Natural Language :: English",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
      ),
      keywords="zope3 mainloop",
      author="gocept gmbh & co. kg",
      author_email="mail@gocept.com",
      url="",
      license="ZPL 2.1",
      package_dir={'': 'src'},
      packages=find_packages('src'),
      include_package_data=True,
      zip_safe=False,
      namespace_packages=['gocept'],
      install_requires=[
          'ZODB3',
          'setuptools',
          'transaction',
          'zope.app.appsetup>=3.6.0',
          'zope.app.component',
          'zope.app.server',
          'zope.app.wsgi',
          'zope.publisher',
          'zope.security',
          'zope.testing',
      ],
      extras_require=dict(
          test=[
              'zope.app.appsetup>=3.6.0',
              'zope.app.testing',
              'zope.app.zcmlfiles',
              'zope.securitypolicy',
               ]),
      entry_points = dict(
        console_scripts =
          ['runexample = gocept.runner.example:example'])
     )
